#!/usr/bin/perl -w

###############################################################################
#
# This file can be distributed and modified freely, as long as this original
# header is not modified.
#
# Author:         Nicolas Hatier    <nicolas dot hatier at niversoft dot com>
# Co-Author:      Bret Miller bret  <dot miller at wcg dot org>
# Website:        http://www.niversoft.com   http://mail.wcg.org/~support/
# Organization:   Niversoft ides logicielles
# Date:           November 2003, April 2004, Mars 2005, November 2006
# Based On:       cgProcessListMessages.wsf by Bret Miller
#
# This code is provided AS IS, without any certification that it will even run
###############################################################################

use strict;
use CLI;
use POSIX ":sys_wait_h";

my $Version = "2.0.3";

# You will need to modify the section below to match your local needs
my %Config =
(
   CGPServer   => "localhost",
   CGPUser     => "postmaster",
   CGPPassword => "password",
   CLIPort     => 106,

   # Process only messages adressed to these lists
   # You must use the EXACT List-ID. Most CGPro List-IDs
   # have the format <listname.servername.ext>
   # See in the headers of a list message to know it's list-id

   ProcessLists =>
   [
      "<list1.example.com>",
      "<list2.example.com>"
   ],

   # "Route" each recipient address and exclude these from recipient substitution

   ExcludeRoutes =>
   [
      '^PIPE.*mypipemodulename'
   ],

   # Remove the List-related headers
   # WARNING: DO NOT ENABLE THIS FEATURE UNLESS YOU KNOW WHAT YOU DO
   # Removing the List headers is a good way to be flagged as a spammer
   RemoveListHeader   => 0,

   # Replace the To: header field with:
   #      0 = Leave it alone
   #      1 = Only the real recipient of the list message
   #      2 = The real recipient PLUS the current contents of the To: header
   #      3 = The current contents of the To: header PLUS the real recipient
   RewriteToHeader    => 2,

   # Add CC: header field with all list recipients.
   # WARNING: DO NOT ENABLE THIS FEATURE UNLESS YOU KNOW WHAT YOU DO
   # USE WITH CARE, YOU'RE REVEALING ALL MEMBERS OF YOUR LISTS TO OTHER MEMBERS

   AddCcHeader       => 0,

   # Add X-ListMember: header field with the real recipient of the list message

   AddXListHeader    => 1,

   # Substitute %RCPT% in the body of the message - put "%RCPT%" somewhere in the message trailer
   # in CGP Mailing List settings, this tag will be replaced by the real recipient of the list message

   SubstituteRCPT    => 1,

   # The following options control whether larger list message are placed on hold
   # and released little-by-little into the submitted folder. The essential logic here is
   # if an incoming message has over $HoldRecipientLimit recipients, then instead of
   # directly submitting the messages, we put them in the $HoldFolder. Then every
   # $HoldPollInterval seconds, we check to see if there are less than $HoldQueueBacklog
   # Messages that haven't yet been accounted for in the CGPro queue and if so, we
   # drop the next $HoldReleaseRate messages in for CGPro to process.
   # Set HoldRecipientLimit to 0 or HoldFolder to "" to disable this feature

   HoldRecipientLimit => 50, # hold messages with at least this many recipients
   HoldFolder         => "./ListHold",
   HoldDelayTime      => 5, #seconds to wait (This must be short since it will delay ending)
   HoldPollInterval   => 60, # seconds
   HoldQueueBacklog   => 20, # difference between "queued messages" and .msg files in the queue folder
   HoldReleaseRate    => 30, # release this many at a time

   # This tag is added to all messages that pass through this processor.
   # You should set the last part of the tag to your server name

   TagHeader       => "X-List-Processed: mail.example.com"
);

sub Log
{
    print "* $_[0]\n";
}

sub Fold
{
   my $line = shift;
   $line =~ s/ +$//g;
   if (length $line > 80)
   {
      $line =~ s/(.{70,}?) +/$1\n  /g;
   }
   return $line;
}

$| = 1;

Log "List Post-Processor v$Version is running";

my $pid;
my $hdir;
my $ReleaseFile;

# Parent process that handles the "hold" queue
if ($pid = fork)
{
   my $DelayLoopsNeeded = $Config{HoldPollInterval} / $Config{HoldDelayTime}; #number of loops we need
   my $DelayLoopCounter = 0;
   my $HoldQueueCount = 0;

   if ($Config{HoldFolder} && $Config{HoldRecipientLimit})
   {
      mkdir $Config{HoldFolder} if (!-d $Config{HoldFolder});
   }

   until(waitpid($pid,WNOHANG) == $pid)
   {
      if ($Config{HoldFolder} && $Config{HoldRecipientLimit})
      {
         $DelayLoopCounter++;
         if ($DelayLoopCounter >= $DelayLoopsNeeded) # time to check
         {
            $DelayLoopCounter = 0; #reset the counter
            opendir $hdir,$Config{HoldFolder};
            my @HoldFolderContents = readdir $hdir;
            closedir $hdir;
            # For some reason, it counts one more than we have
            $HoldQueueCount = $#HoldFolderContents - 1;
            if ($HoldQueueCount > 0) #we have waiting messages
            {
               print "* Held messages: $HoldQueueCount\n";
               opendir $hdir,"Queue";
               my @QueueFiles = readdir $hdir;
               closedir $hdir;
               my $cli = new CGP::CLI
                        ({
                           PeerAddr => $Config{CGPServer},
                           PeerPort => $Config{CLIPort},
                           login => $Config{CGPUser},
                           password => $Config{CGPPassword}
                        })
                  || die "*** Can't login to CGPro CLI: ".$CGP::ERR_STRING."\n";

               my $QueueCount = $cli->GetSNMPElement('1.3.6.1.4.1.5678.2.1.2.1');
               $cli->Logout;

               if (($#QueueFiles-$QueueCount) <= $Config{HoldQueueBacklog})
               {
                  my $ReleaseCount = $Config{HoldReleaseRate};
                  do
                  {
                     $ReleaseFile = pop @HoldFolderContents;
                     if (defined $ReleaseFile)
                     {
                        rename "$Config{HoldFolder}/$ReleaseFile","Submitted/$ReleaseFile";
                     }
                     $ReleaseCount--;
                  }
                  while ($ReleaseCount && defined $ReleaseFile);
               }
               else
               {
                  #print "* $QueueCount queued messages.\n";
                  print "* Queue backlog: ".($#QueueFiles-$QueueCount).".\n";
               }
            }
         }
      }
      sleep $Config{HoldDelayTime}; # wait a few seconds
   }

} # End of Parent "hold" queue function
#
# Child process that handles the list post-processing
else
{
   my $UniqueID = 0;
   my $processrecipient;
   mkdir "Submitted" if (!-d "Submitted");
   while (<>)
   {
      my @line = split(/ /, $_);
      $line[0] =~ s/[\r\n]//g;
      print "1 OK\n"             and next unless (length $line[0]);
      print "$line[0] OK\n"      and next unless (defined $line[1]);
      print "$line[0] OK\n"      and next if ($line[1] =~ /^quit$/i);
      print "$line[0] INTF 3\n"  and next if ($line[1] =~ /^intf$/i);
      print "$line[0] OK\n"      and next if ($line[1] =~ /^key$/i);
      print "$line[0] FAILURE\n" and next if ($line[1] !~ /^file$/i);

      Log "Nothing To Do!\n$line[0] OK" and next if (!$Config{AddXListHeader} && !$Config{RewriteToHeader} & !$Config{SubstituteRCPT});

      $line[2] =~ s|\\|/|g; # Getting the filepath
      $line[2] =~ s/[\r\n]//g;

      if (!open(MSG, $line[2]))
      {
         close(MSG);
         Log "File Not Found $line[2]\n$line[0] OK";
         next;
      }

      #Log "Analyzing $line[2]";
      my $MsgID = "(no listid)";
      $MsgID = $1 if ($line[2] =~ m|Queue/(.*)\.msg|i);

      my $Sender = "<>";
      my %Recipients;
      my $Subject = "";
      my $ListID = "";
      my $Process = 1;

      while (<MSG>) #process CGP message info
      {
         s/[\r\n]//g;
         last if($_ eq '');
         if(/^(\w) (\w).+<(.+)>/)
         {
            $Sender = $3 if   ($1 eq 'P');
            $Recipients{$3} = 1 if (($1 eq 'R') and ($2 eq 'W'));
         }
      }

      #Log "Fetched Envelope";

      my $HeaderPos = tell MSG;
      while (<MSG>) #process header
      {
         s/[\r\n]//g;
         last if($_ eq '');

         $Subject = $1 if /^Subject ?: ?(.*)$/i;
         $ListID  = $1 if /^List-ID ?: ?(.*)$/i;
         $Process =  0 if /^$Config{TagHeader}$/i;
      }

      unless ($Process-- and $ListID)
      {
         close(MSG);
         print "$line[0] OK\n";
         next;
      }

      foreach (@{$Config{ProcessLists}})
      {
         if (/$ListID/i)
         {
            $Process = 1;
            last;
         }
      }

      unless($Process)
      {
         close(MSG);
         print "$line[0] OK\n";
         next;
      }

      #Log "Beginning list post-processing";
      Log "$line[0] $MsgID processing list $ListID";
      Log "$line[0] $MsgID Subject: $Subject";


      my $cli = new CGP::CLI
                  ({
                    PeerAddr => $Config{CGPServer},
                    PeerPort => $Config{CLIPort},
                    login    => $Config{CGPUser},
                    password => $Config{CGPPassword}
                 })
        || die "*** Can't login to CGPro CLI: ".$CGP::ERR_STRING."\n";

      my @Recipients = keys %Recipients;
      my $FolderToUse =
         ($Config{HoldFolder} && $Config{HoldRecipientLimit} && ($#Recipients >= $Config{HoldRecipientLimit}))
         ? $Config{HoldFolder} : "Submitted";
      Log "$line[0] $MsgID using $FolderToUse folder for this message";

      foreach my $recipient (@Recipients)
      {
         my $CCRecipients;
         if ($Config{AddCcHeader})
         {
            delete $Recipients{$recipient} if ($Config{RewriteToHeader});
            $CCRecipients = join(", ", keys %Recipients);
            $Recipients{$recipient} = 1;
         }

         my $filename = "listsub_".time.($UniqueID++);
         open FILE, ">Submitted/$filename.tmp";
         print FILE "Return-Path: $Sender\n";
         print FILE "Envelope-To: $recipient\n";
         print FILE "$Config{TagHeader}\n";
         print FILE "X-ListMember: $recipient\n" if($Config{AddXListHeader});

         my $routedata = $cli->Route("$recipient",'mail');
         my $routeresult = $routedata->[0].$routedata->[1].$routedata->[2];

         $processrecipient = 1;
         my $excludedroute;
         foreach $excludedroute (@{$Config{ExcludeRoutes}})
         {
            if ($routeresult =~ /$excludedroute/i)
            {
               $processrecipient = 0;
               last;
            }
         }

         seek MSG, $HeaderPos, 0;
         my @Headers = ();
         my $HasCc = 0;
         while (<MSG>) #process header
         {
            s/[\r\n]//g;
            if (/^[ \t]+(.*)/)
            {
               $Headers[scalar @Headers - 1] = $Headers[scalar @Headers - 1] . " $1";
            }
            else
            {
              push @Headers, $_;
              $HasCc = $HasCc || /^Cc ?:/i;
            }
            last if($_ eq '');
         }
         my $FoundCC = 0;
         foreach (@Headers)
         {
            if ($Config{RemoveListHeader}
            and (($_ =~ /^X-ListServer ?:/)
             or  ($_ =~ /^List-Unsubscribe ?:/)
             or  ($_ =~ /^List-ID ?:/)
             or  ($_ =~ /^List-Archive ?:/)
             or  ($_ =~ /^Precedence ?:/)))
            {
                next;
            }
            if (/^To ?\: ?/i && ($Config{RewriteToHeader} eq 1))
            {
               print FILE Fold("To: $recipient") . "\n";
            }
            elsif (/^To ?\: ?/i && !/$recipient/i && ($Config{RewriteToHeader} eq 2))
            {
              s/^To ?\: /To: $recipient,/i;
              print FILE Fold($_) . "\n";
            }
            elsif (/^To ?\: ?/i && !/$recipient/i && ($Config{RewriteToHeader} eq 3))
            {
               print FILE Fold("$_, $recipient") . "\n";
            }
            elsif (/^Cc ?\: ?(.*)$/i && ($Config{AddCcHeader}))
            {
               print FILE Fold("Cc: $1, $CCRecipients") . "\n";
               $FoundCC = 1;
            }
            else
            {
               print FILE Fold("$_") . "\n";
            }

            if (!$HasCc && ($Config{AddCcHeader}) && /^To ?\:/i)
            {
               print FILE Fold("Cc: $CCRecipients") . "\n";
               $HasCc = 1;
            }
         }

         if ($Config{SubstituteRCPT} && $processrecipient)
         {
            while (<MSG>)
            {
               s/\%RCPT\%/$recipient/g;
               s/\%25RCPT\%25/$recipient/g;
               print FILE $_;
            }
         }
         else
         {
            print FILE $_ while (<MSG>);
         }
         close FILE;
         rename "Submitted/$filename.tmp","$FolderToUse/$filename.sub";

         Log "$line[0] $MsgID submitted $filename for $recipient";
      }
      close MSG;
      $cli->Logout;
      print "$line[0] DISCARD\n";

      # Just to ensure that STDOUT is (really) correctly flushed
      open STDOUT, ">&STDOUT";
   }
} # End of child list post-processor function
